#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A function to normalise projection data using the generated flat fields, Poisson
noise is applied to flat fields and data

The errors in the flat fields will be propagated into prokjection data after
normalisation. This can have a serious negative effect on the reconstructed images 
in terms of the quality of reconstruction (artifacts presence). 

This modelling, however, is more realistic and recommended if algorithms are 
tested on robustness towards artifacts and errors in data

@author: Daniil Kazantsev
"""
import random
import numpy as np
from tomophantom.supp.artifacts import noise

def normaliser_sim(projData3D, flatSIM, sigma_flats = 5000, flux_intensity = 30000):
    """
    projData3D - 3D projection data (noiseless) [DetectorsDimV, Proj_angles, DetectorsDimH]
    maxthickness - a value in pixels which controls the width of stripes
    sigma_flats - a noise level (Poisson) in flats
    flux_intensity  - controls the level of Posson noise applied to projection data
    """
    [DetectorsDimV, Proj_angles, DetectorsDimH] = np.shape(projData3D)
    [flatsnum, DetectorsDimV_f, DetectorsDimH_f] = np.shape(flatSIM)
    if (DetectorsDimV != DetectorsDimV_f):
        raise("The size of the vertical detector for data and the flat field is different ")
    if (DetectorsDimH != DetectorsDimH_f):
        raise("The size of the horizontal detector for data and the flat field is different ")
    # add noise to the stack of images (flats)
    flat_all_noise = noise(sinogram = flatSIM, sigma=sigma_flats, noisetype='Poisson', seed = None, prelog = None)
    flat_average_noise = np.average(flat_all_noise,0) # calculate average of all flats

    nonzeroInd = np.where(flat_average_noise != 0) # nonzero data
    zeroInd = np.where(flat_average_noise == 0) # zero data
    projData3D_norm = np.zeros(np.shape(projData3D),dtype='float32')
    projData3D_raw = np.zeros(np.shape(projData3D),dtype='int64')
    
    for x in range(0,Proj_angles):
        proj2D = projData3D[:,x,:]
        norm_proj = np.zeros(np.shape(proj2D))
        randflatind = random.randint(0,flatsnum-1) # generate random flat number
        proj_flat = flatSIM[randflatind,:,:]*proj2D
        norm_proj[nonzeroInd]  = proj_flat[nonzeroInd]/flat_average_noise[nonzeroInd]
        [norm_proj_noisy, proj_raw] = noise(sinogram = norm_proj, sigma=flux_intensity, noisetype='Poisson', seed = None, prelog = True)
        norm_proj_noisy[zeroInd] = 1e-13
        projData3D_norm[:,x,:] = np.float32(norm_proj_noisy)
        projData3D_raw[:,x,:] = proj_raw
    return [projData3D_norm, projData3D_raw]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A function to generate fake flat field images for 3D projection data normalisation

@author: Daniil Kazantsev
"""

from scipy.special import spherical_yn
from scipy.special import y1
from scipy.ndimage.filters import gaussian_filter
import random
import numpy as np
# import matplotlib.pyplot as plt

def flats(DetectorsDimV, DetectorsDimH, maxheight, maxthickness, sigma_noise, sigmasmooth, flatsnum):
    """
    maxheight - a value between (0,1] which controls the height of stripes
    maxthickness - a value in pixels which controls the width of stripes
    sigma_noise - a noise level (Gaussian) which is added to the sripe image
    sigmasmooth - a smoothing parameter for a stripe image with noise (1,3,5,7...)
    flatsnum - a number of flats to generate
    """
    
    flatfield = np.zeros((DetectorsDimV,DetectorsDimH))
    flat_combined3D = np.zeros((flatsnum,DetectorsDimV,DetectorsDimH))
    
    # using spherical Bessel functions
    func = spherical_yn(1, np.linspace(3,15,DetectorsDimV,dtype='float32'))
    func = func + abs(np.min(func))
    func2 = y1(np.linspace(15,5,DetectorsDimH,dtype='float32'))
    func2 = func2 + abs(np.min(func2))
    
    for i in range(0,DetectorsDimV):
        flatfield[i,:] = func2
    
    for i in range(0,DetectorsDimH):
        flatfield[:,i] += func
    
    # Adding stripes of varying vertical & horizontal sizes
    maxheight_par = round(maxheight*DetectorsDimV)
    max_intensity = np.max(flatfield)
    flatstripes = np.zeros(np.shape(flatfield))
    
    for x in range(0,DetectorsDimH):
        randind = random.randint(0,DetectorsDimV) # generate random index (vertically)
        randthickness = random.randint(0,maxthickness) #generate random thickness
        randheight = random.randint(0,maxheight_par) #generate random height
        randintens = random.uniform(0.1, 2.0) # generate random multiplier
        intensity = max_intensity*randintens
        
        for x1 in range(-randheight,randheight):
            if (((randind+x1) >= 0) and ((randind+x1) < DetectorsDimV)):
                for x2 in range(-randthickness,randthickness):
                    if (((x+x2) >= 0) and ((x+x2) < DetectorsDimH)):
                        flatstripes[randind+x1,x+x2] += intensity
    
    for i in range(0,flatsnum):
        # adding noise and normalise
        flatstripes2 = flatstripes.copy()
        flatstripes2 += np.random.normal(loc = 0.00 ,scale = sigma_noise, size = np.shape(flatstripes2))
        flatstripes2 /= np.max(flatstripes2)
        flatstripes2 += abs(np.min(flatstripes2))
        
        #blur the result and add to the initial image with bessel background
        blurred_flatstripes = gaussian_filter(flatstripes2, sigma=sigmasmooth)
        flat_combined = flatfield + blurred_flatstripes
        flat_combined /= np.max(flat_combined)
        flat_combined3D[i,:,:] = flat_combined
    return flat_combined3D


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Various methods to simulate noise and artifacts for sinograms (2D) and for 3D-projection data
 
What can be simulated: 
-- noise (Poisson or Gaussian)
-- zingers (result in streaks in the reconstruction)
-- stripes (result in rings in the reconstruction)
---- variable intensity
---- partial
-- shifts - data misalignment (result in blur in the reconstruction)

Note that the TomoPhantom package is released under Apache License, Version 2.0
@author: Daniil Kazantsev
"""

def _Artifacts_(sinogram, 
                _noise_, # noise related dictionary
                _zingers_, # zingers dictionary
                _stripes_, # stripes dictionary
                _sinoshifts_ #sinoshifts
                ):
    ####### NOISE ########
    if ('type' not in _noise_):
        _noise_['type'] = 'Poisson'
    if ((_noise_['type'] != 'Gaussian') and (_noise_['type'] != 'Poisson')):
        print("Unknown noise type, select Gaussian or Poisson, now set to Poisson")
        _noise_['type'] = 'Poisson'
    if ('sigma' not in _noise_):
        # photon intensity (Poisson) or variance for Gaussian 
        _noise_['sigma'] = 10000
    if ('seed' not in _noise_):
        # seeds for noise (integers), None means random generation
        _noise_['seed'] = None
    if ('prelog' not in _noise_):
        # set to True in order to get the prelog (raw) data 
        _noise_['prelog'] = None
    ####### ZINGERS ########
    if ('percentage' not in _zingers_):
        # # percentage the amount of zingers to be added to the data
        _zingers_['percentage'] = None
    if ('modulus' not in _zingers_):
        # modulus to control the amount of 4/6 pixel clusters to be added
        _zingers_['modulus'] = 10
    ####### STRIPES ########
    if ('percentage' not in _stripes_):
        # percentage defines the amount of stripes in the data
        _stripes_['percentage'] = None
    if ('maxthickness' not in _stripes_):
        # maxthickness defines the maximal thickness of a stripe
        _stripes_['maxthickness'] = 1.0
    if ('intensity' not in _stripes_):
        # controls the intensity levels of stripes
        _stripes_['intensity'] = 0.1
    if ('type' not in _stripes_):
        # stripe types can be 'partial' or 'full'
        _stripes_['type'] = 'full'
    # variability multiplier to incorporate change of intensity in the stripe
    if ('variability' not in _stripes_):
        _stripes_['variability'] = 0.0
    ####### SINOSHIFTS ########
    if ('maxamplitude' not in _sinoshifts_):
        # maxamplitude (in int pixels) defines the maximal amplitude of each angular deviation
        _sinoshifts_['maxamplitude'] = None
    'Applying artifacts and noise to the data'
    # ZINGERS
    if _zingers_['percentage'] is not None:
        sino_artifacts = zingers(sinogram=sinogram, percentage=_zingers_['percentage'], modulus = _zingers_['modulus'])
        print("Zingers have been added to the data.")
    else:
        sino_artifacts = sinogram
    # STRIPES
    if _stripes_['percentage'] is not None:
        sino_artifacts = stripes(sinogram=sino_artifacts,\
                                 percentage=_stripes_['percentage'],\
                                 maxthickness = _stripes_['maxthickness'],\
                                 intensity_thresh = _stripes_['intensity'],\
                                 stripe_type = _stripes_['type'],\
                                 variability = _stripes_['variability'])
        print("Stripes have been added to the data.")
    # SINOSHIFTS
    if _sinoshifts_['maxamplitude'] is not None:
        [sino_artifacts, shifts] = sinoshifts(sinogram=sino_artifacts, maxamplitude = _sinoshifts_['maxamplitude'])
        print("Sinogram shifts have been added to the data.")
    # NOISE
    sino_artifacts = noise(sinogram=sino_artifacts, sigma=_noise_['sigma'], noisetype = _noise_['type'], seed = _noise_['seed'], prelog=_noise_['prelog'])
    print("{} noise have been added to the data.".format(_noise_['type']))
    
    if _sinoshifts_['maxamplitude'] is not None:
        return [sino_artifacts,shifts]
    else:
        return sino_artifacts

def stripes(sinogram, percentage, maxthickness, intensity_thresh, stripe_type, variability):
    # Function to add stripes (constant offsets) to sinogram which results in rings in the 
    # reconstructed image
    # - percentage defines the amount of stripes in the data
    # - maxthickness defines the maximal thickness of a stripe
    # - stripe_type can be 'partial' or 'full'
    # - variability multiplier to incorporate change of intensity in the stripe
    import numpy as np
    import random
    if (sinogram.ndim == 2):
        (anglesDim, DetectorsDimH) = np.shape(sinogram)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(sinogram)
    if 0 < percentage <= 100:
        pass
    else:
        raise ("percentage must be larger than zero but smaller than 100")
    if 0 <= maxthickness <= 10:
        pass
    else:
        raise ("maximum thickness must be in [0,10] range")
    if ((stripe_type != 'partial')):
        stripe_type = 'full'
    sino_stripes = sinogram.copy()
    max_intensity = np.max(sino_stripes)
    range_detect = int((np.float32(DetectorsDimH))*(np.float32(percentage)/100.0))
    if (sinogram.ndim == 2):
        for x in range(range_detect):
            for mm in range(0,20):
                randind = random.randint(0,DetectorsDimH-1) # generate random index
                if (sino_stripes[0,randind] != 0.0):
                    break
            if (stripe_type == 'partial'):
                randind_ang1 =  random.randint(0,anglesDim) 
                randind_ang2 =  random.randint(0,anglesDim)
            else:
                randind_ang1 = 0
                randind_ang2 = anglesDim
            randthickness = random.randint(0,maxthickness) #generate random thickness
            randintens = random.uniform(-1.0, 0.5) # generate random multiplier
            intensity = max_intensity*randintens*intensity_thresh
            if ((randind > 0+randthickness) & (randind < DetectorsDimH-randthickness)):
                for x1 in range(-randthickness,randthickness+1):
                    if (variability != 0.0):
                        intensity_off = variability*max_intensity
                    else:
                        intensity_off = 0.0
                    for ll in range(randind_ang1,randind_ang2):
                        sino_stripes[ll,randind+x1] += intensity + intensity_off
                        intensity_off += [-1,1][random.randrange(2)]*variability*max_intensity
                        #sino_stripes[randind_ang1:randind_ang2,randind+x1] += intensity
    else:
        for j in range(DetectorsDimV):
            for x in range(range_detect):
                for mm in range(0,20):
                    randind = random.randint(0,DetectorsDimH-1) # generate random index
                    if (sino_stripes[j,0,randind] != 0.0):
                        break
                if (stripe_type == 'partial'):
                    randind_ang1 =  random.randint(0,anglesDim) 
                    randind_ang2 =  random.randint(0,anglesDim)
                else:
                    randind_ang1 = 0
                    randind_ang2 = anglesDim
                randthickness = random.randint(0,maxthickness) #generate random thickness
                randintens = random.uniform(-1, 0.5) # generate random multiplier
                intensity = max_intensity*randintens*intensity_thresh
                if ((randind > 0+randthickness) & (randind < DetectorsDimH-randthickness)):
                    for x1 in range(-randthickness,randthickness+1):
                        if (variability != 0.0):
                            intensity_off = variability*max_intensity
                        else:
                            intensity_off = 0.0
                        for ll in range(randind_ang1,randind_ang2):
                            sino_stripes[j,ll,randind+x1] += intensity + intensity_off
                            intensity_off += [-1,1][random.randrange(2)]*variability*max_intensity
    return sino_stripes

def zingers(sinogram, percentage, modulus):
    # adding zingers (zero single pixels or small 4 pixels clusters) to data or 6 voxels to 3D projection data
    # percentage = 0.5, #- percentage - the amount of zingers to be added to the data
    # modulus = 10 # modulus to control the amount of 4/6 pixel clusters to be added
    import numpy as np
    import random
    if (sinogram.ndim == 2):
        (anglesDim, DetectorsDimH) = np.shape(sinogram)
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(sinogram)
    if 0.0 < percentage <= 100.0:
        pass
    else:
        raise ("percentage must be larger than zero but smaller than 100")
    if (modulus > 0):
        pass
    else:
        raise ("Modulus integer must be positive")
    sino_zingers = sinogram.copy()
    length_sino = np.size(sino_zingers)
    num_values = int((length_sino)*(np.float32(percentage)/100.0))
    sino_zingers_fl = sino_zingers.flatten()
    for x in range(num_values):
        randind = random.randint(0,length_sino) # generate random index 
        sino_zingers_fl[randind] = 0
        if ((x % int(modulus)) == 0):
            if (sinogram.ndim == 2):
                if ((randind > DetectorsDimH) & (randind < length_sino-DetectorsDimH)):
                    sino_zingers_fl[randind+1] = 0
                    sino_zingers_fl[randind-1] = 0
                    sino_zingers_fl[randind+DetectorsDimH] = 0
                    sino_zingers_fl[randind-DetectorsDimH] = 0
            else:
                if ((randind > DetectorsDimH*DetectorsDimV) & (randind < length_sino-DetectorsDimH*DetectorsDimV)):
                    sino_zingers_fl[randind+1] = 0
                    sino_zingers_fl[randind-1] = 0
                    sino_zingers_fl[randind+DetectorsDimH] = 0
                    sino_zingers_fl[randind-DetectorsDimH] = 0
                    sino_zingers_fl[randind+DetectorsDimH*DetectorsDimV] = 0
                    sino_zingers_fl[randind-DetectorsDimH*DetectorsDimV] = 0
    sino_zingers[:] = sino_zingers_fl.reshape(sino_zingers.shape)
    return sino_zingers

def noise(sinogram, sigma, noisetype, seed, prelog):
    # Adding random noise to data
    # noisetype = None, # 'Gaussian', 'Poisson' or None
    # sigma = 10000, # photon intensity (Poisson) or variance for Gaussian 
    # seed = 0, # seeds for noise
    # prelog: None or True (get the raw pre-log data) 
    import numpy as np
    sino_noisy = sinogram.copy()
    if noisetype == 'Gaussian':
        # add normal Gaussian noise
        if seed is not None:
            np.random.seed(int(seed))
        sino_noisy += np.random.normal(loc = 0.0, scale = sigma, size = np.shape(sino_noisy))
        sino_noisy[sino_noisy<0] = 0
    elif noisetype == 'Poisson':
        # add Poisson noise
        maxSino = np.max(sinogram)
        if maxSino > 0:
            sino_noisy = sinogram/maxSino
            dataExp = sigma*np.exp(-sino_noisy)  # noiseless raw data
            sino_raw = np.random.poisson(dataExp) #adding Poisson noise
            div_res = np.float32(sino_raw)/np.max(sino_raw)
            sino_noisy = -np.log(div_res)*maxSino # log corrected data -> sinogram
            sino_noisy[sino_noisy<0] = 0
    else:
        print ("Select 'Gaussian' or 'Poisson' for noise type")
    if prelog is True:
        return [sino_noisy,sino_raw]
    else:
        return sino_noisy

def sinoshifts(sinogram, maxamplitude):
    # A function to add random shifts to sinogram rows (an offset for each angular position)
    # maxamplitude (integer pixels) defines the maximal amplitude of each angular deviation
    import numpy as np
    import random
    if (sinogram.ndim == 2):
        (anglesDim, DetectorsDimH) = np.shape(sinogram)
        shifts = np.zeros(anglesDim, dtype='int8') # the vector of shifts
    else:
        (DetectorsDimV, anglesDim, DetectorsDimH) = np.shape(sinogram)
        shifts = np.zeros([anglesDim,2], dtype='int8') # the 2D vector of shifts

    sino_shifts = np.zeros(np.shape(sinogram),dtype='float32')
    non = lambda s: s if s<0 else None
    mom = lambda s: max(0,s)
    for x in range(anglesDim):
        rand_shift = random.randint(-maxamplitude,maxamplitude)  #generate random shift
        if (sinogram.ndim == 2):
            shifts[x] = rand_shift
            projection = sinogram[x,:] # extract 1D projection
            projection_shift = np.zeros(np.shape(projection),dtype='float32')
            projection_shift[mom(rand_shift):non(rand_shift)] = projection[mom(-rand_shift):non(-rand_shift)]
            sino_shifts[x,:] = projection_shift
        else:
            rand_shift2 = random.randint(-maxamplitude,maxamplitude)  #generate random shift
            shifts[x,0] = rand_shift
            shifts[x,1] = rand_shift2
            projection2D = sinogram[:,x,:] # extract 2D projection
            projection2D_shift = np.zeros(np.shape(projection2D),dtype='float32')
            projection2D_shift[mom(rand_shift):non(rand_shift), mom(rand_shift2):non(rand_shift2)] = projection2D[mom(-rand_shift):non(-rand_shift),mom(-rand_shift2):non(-rand_shift2)]
            sino_shifts[:,x,:] = projection2D_shift
    return [sino_shifts,shifts]
